
#ifndef PLS_ABSTRACT_TASK_H
#define PLS_ABSTRACT_TASK_H

#define PLS_UNIQUE_ID __COUNTER__

#include "pls/internal/base/spin_lock.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            class abstract_task {
            public:
                struct id {
                    uint32_t id_;
                    bool auto_generated_;

                    explicit id(uint32_t id, bool auto_generated=false): id_{id}, auto_generated_{auto_generated} {};

                    bool operator==(const abstract_task::id& other) const {
                        return id_ == other.id_ && auto_generated_ == other.auto_generated_;
                    }
                };

            private:
                int depth_;
                abstract_task::id unique_id_;
                abstract_task* child_task_;

            public:
                abstract_task(const int depth, const abstract_task::id& unique_id):
                    depth_{depth},
                    unique_id_{unique_id},
                    child_task_{nullptr} {}

                virtual void execute() = 0;
                void set_child(abstract_task* child_task) { child_task_ = child_task; }
                abstract_task* child() { return child_task_; }

                void set_depth(int depth) { depth_ = depth; }
                int depth() const { return depth_; }
                id unique_id() const { return unique_id_; }
            protected:
                virtual bool internal_stealing(abstract_task* other_task) = 0;
                virtual bool split_task(base::spin_lock* lock) = 0;

                bool steal_work();
            };
        }
    }
}

#endif //PLS_ABSTRACT_TASK_H
