#include "pls/internal/base/aligned_stack.h"
#include "pls/internal/base/thread.h"

#include "thread_state.h"

#ifndef PLS_SCHEDULER_MEMORY_H
#define PLS_SCHEDULER_MEMORY_H

namespace pls {
    namespace internal {
        namespace scheduling {
            void worker_routine();
            using scheduler_thread = base::thread<decltype(&worker_routine), thread_state>;

            class scheduler_memory {
            public:
                virtual size_t max_threads() = 0;
                virtual thread_state* thread_state_for(size_t id) = 0;
                virtual scheduler_thread* thread_for(size_t id) = 0;
                virtual base::aligned_stack* task_stack_for(size_t id) = 0;
            };

            template<size_t MAX_THREADS, size_t TASK_STACK_SIZE>
            class static_scheduler_memory: public scheduler_memory {
                std::array<scheduler_thread, MAX_THREADS> threads_;
                std::array<thread_state, MAX_THREADS> thread_states_;
                std::array<std::array<char, TASK_STACK_SIZE>, MAX_THREADS> task_stacks_memory_;
                std::array<base::aligned_stack, MAX_THREADS> task_stacks_;

            public:
                static_scheduler_memory() {
                    for (size_t i = 0; i < MAX_THREADS; i++) {
                        task_stacks_[i] = base::aligned_stack(task_stacks_memory_[i].data(), TASK_STACK_SIZE);
                    }
                }

                size_t max_threads() override { return MAX_THREADS; }
                thread_state* thread_state_for(size_t id) override { return &thread_states_[id]; }
                scheduler_thread* thread_for(size_t id) override { return &threads_[id]; }
                base::aligned_stack* task_stack_for(size_t id) override { return &task_stacks_[id]; }
            };

            class malloc_scheduler_memory: public scheduler_memory {
                size_t num_threads_;

                scheduler_thread* threads_;
                thread_state* thread_states_;
                char** task_stacks_memory_;
                base::aligned_stack* task_stacks_;
            public:
                explicit malloc_scheduler_memory(size_t num_threads, size_t memory_per_stack = 2 << 16);
                ~malloc_scheduler_memory();

                size_t max_threads() override { return num_threads_; }
                thread_state* thread_state_for(size_t id) override { return &thread_states_[id]; }
                scheduler_thread* thread_for(size_t id) override { return &threads_[id]; }
                base::aligned_stack* task_stack_for(size_t id) override { return &task_stacks_[id]; }
            };
        }
    }
}

#endif //PLS_SCHEDULER_MEMORY_H
